# Copyright 2019 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import logging

# This is exported by virtinst/__init__.py
log = logging.getLogger("virtinst")
